#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Designed Nov. 2007 by Fredo6

# This software is provided as an example of using the Ruby interface to SketchUp.

# Permission to use, copy, modify, and distribute this software for 
# any purpose and without fee is hereby granted, provided that the above
# copyright notice appear in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name		:   LibProtractorTool.rb
# Type		:   Utility LIbrary as a Tool
# Description	:   Feature a Protractor tool for getting inputs of an origin, a plane,and 2 vectors
# Menu Item	:   none
# Context Menu	:   none
# Usage		:   See Tutorial 
# Date		:   15 Dec 2007
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************

require 'sketchup.rb'
require 'libtraductor.rb'		# for language translation

module LibProtractorTool

STATE_ORIGIN = 0
STATE_DIRECTION = 1
STATE_ROTATION = 2
STATE_EXECUTION = 3

MSG_Degree = "Degree |FR| Degr\"
MSG_Direction = "Direction |FR| Direction"
MSG_Rotation = "Rotation |FR| Rotation"
MSG_PlaneAngle = "Plane Angle |FR| Angle du plan"
MSG_Input_Origin = ["Select Origin and Plane (Shift to lock plane)",
                    "|FR| S\lectionner l'origine et le plan (Maj pour verrouiller plan)"]
MSG_Input_Direction = ["Draw reference direction (angle value in VCB)",
                       "|FR| Tracer la direction de r\f\rence (valeur de l'angle dans la VCB)"]
MSG_Input_Rotation = ["Draw direction of rotation (angle value in VCB) - Double Click to launch operation: %1",
                    "|FR| Tracer la direction de rotation (valeur de l'angle dans la VCB) - Double cliquez pour ex\cuter l'op\ration : %1"]
MSG_Input_Execution = ["Double Click or Enter to launch operation '%1' (or adjust angle value in VCB)",
                    "|FR| Double cliquez ou Enter pour ex\cuter l'op\ration '%1' (ou valeur de l'angle dans la VCB)"]
					

class ProtractorTool

Direction = Struct.new("Direction", :ip, :pt, :vec, :angle, :angle_imposed, :freedom) 

def direction_create(freedom)
	dir = Direction.new()
	dir.ip = Sketchup::InputPoint.new
	dir.freedom = freedom
	dir
end

def initialize(caller_class, cursorpath=nil, opname="")
	Traductor.load_translation LibProtractorTool, /MSG_/, binding, "@msg_"
	@caller = caller_class
	@cursorpath = cursorpath
	@opname = opname
	@idcursor = 0
	@ip_origin = Sketchup::InputPoint.new
	@origin = ORIGIN
	@axis_def = Y_AXIS.reverse
	@angle_def = 0.0
	@normal_def = @axis_def
	@normal = @axis_def
	@pt_ref = ORIGIN
	@bigcircle = []
	@graduations = []
	@inner_up = []
	@inner_down = []
	@gray_color = Sketchup::Color.new "Gray"
	@face_color = @gray_color
	n = 24
	pi = Math::PI
	radius = 200.cm
	fact_grad = 0.9
	fact_inner = 0.75
	limit = Math::sin(pi / 12)
	
	#defining the Protractor
	for i in 0..n
		angle = 2 * pi * i / n
		cosinus = Math::cos angle
		sinus = Math::sin angle
		pt_ext = Geom::Point3d.new radius * cosinus, radius * sinus, 0
		@bigcircle.push pt_ext
		pt_int = Geom::Point3d.new radius * fact_grad * cosinus, radius * fact_grad * sinus, 0
		@graduations.push pt_int
		@graduations.push pt_ext
		if (sinus.abs >= limit)
			pt_ext = Geom::Point3d.new radius * fact_inner * cosinus, radius * fact_inner * sinus, 0
			if (sinus > 0.0)
				@inner_up.push pt_ext
			else
				@inner_down.push pt_ext
			end	
		end
	end
	@inner_up.push @inner_up[0]
	@inner_down.push @inner_down[0]
	
	#Precomputing the geometry of the view in pixel coordinates
	view = Sketchup.active_model.active_view
	@vpwidth = view.vpwidth
	@vpheight = view.vpheight
	@ptbox = []
	@ptbox[0] = Geom::Point3d.new(0, 0, 0)
	@ptbox[1] = Geom::Point3d.new(0, 0, 0)
	
	#Initializing the structures to hold the direction and rotation transformations
	@dir_base = direction_create 1
	@dir_rot = direction_create 0
	
	self.compute_origin
	
	if (@cursorpath)
		f = Sketchup.find_support_file @cursorpath, "Plugins"
		@idcursor = UI::create_cursor f, 10, 10
	end	
end

def activate
	set_state STATE_ORIGIN
end

def deactivate(view)
	view.invalidate
end

def onSetCursor
	UI::set_cursor @idcursor if (@state == STATE_EXECUTION && @idcursor != 0)
end

def call_execute
	return unless @caller
	hmeth = @caller.method 'protractor_execute'
	if (hmeth == nil || hmeth.arity != 3)
		UI.beep
	else
		begin
			if hmeth.call @origin, @dir_base.vec, @dir_rot.vec
				Sketchup.active_model.select_tool nil 
				return
			end	
		#rescue			
		end	
	end
	set_state STATE_ROTATION
end

def onLButtonDoubleClick(flags, x, y, view)
	if @state == STATE_EXECUTION
		call_execute
	else
		UI.beep
	end
end

def onKeyDown(key, rpt, flags, view)
	key = Traductor.check_key key, flags, false
	puts "Key down = #{key}"
	case key
	when VK_UP
		newnormal = Z_AXIS
	when VK_RIGHT
		newnormal = X_AXIS	
	when VK_LEFT
		newnormal = Y_AXIS.reverse	
	when VK_DOWN
		newnormal = @axis_def
	else
		return false
	end
	@axis_def = newnormal
	if (@normal == @normal_def)
		@normal = @axis_def
		self.compute_origin
		view.invalidate
	end	
	@normal_def = @axis_def
	@angle_def = 0.0
	info_angle @angle_def
end

def onKeyUp(key, rpt, flags, view)
	key = Traductor.check_key key, flags, true
	case key
	when 13			#Return key
		unless @usertext
			return UI.beep if @state < STATE_ROTATION
			call_execute
		end	
		@usertext = false
	else
		return false
	end	
	true
end

def onUserText(text, view)
	@usertext = true
	angle = parse_angle text
	return UI.beep unless angle
	
	case @state
	when STATE_ORIGIN
		@angle_def = angle
		@normal_def = Geom::Transformation.rotation(@origin, @axesD[0], @angle_def) * @axis_def
		@normal = @normal_def
		self.compute_origin
	when STATE_DIRECTION
		direction_set_angle @dir_base, angle
	else	
		direction_set_angle @dir_rot, angle
	end
	view.invalidate
	info_angle angle
	set_state @state + 1 unless @state == STATE_ORIGIN
end

def parse_angle (text)
	text = text.strip
	if text =~ /^=/
		st = $'
		puts "formula #{text} --> #{st}"
		begin
			if st =~ /d$/i
				sd = formula_float $`
				puts "formula degree #{text} --> #{sd}"
				dangle = Kernel::eval sd
			elsif st =~ /r$/i
				sd = formula_float $`
				angle = (Kernel::eval sd)
				dangle = angle.to_f * 180 / Math::PI
				puts "formula radian #{text} --> #{sd}"
			else
				st = formula_float(st)
				puts "formula tangente #{text} --> st = #{st}"
				tg = Kernel::eval st
				dangle = (tg == 0.0) ? 0.0 : ((Math.atan2(tg, 1)) * 180 / Math::PI)
				puts "formula tangente #{text} --> #{st}"
			end	
		rescue
			UI.beep
			return false
		end
	elsif text =~ /^(d|.)/
		dangle = text.to_f
		puts "numeric #{text} --> #{dangle}"
	else
		UI.beep
		return false
	end	
	puts "angle = #{dangle}"
	dangle = dangle.modulo 360.0
	angle = dangle * Math::PI / 180
end

def formula_float(s)
	s2 = s.gsub(/\.\d/) { |m| ($`[-1..-1] =~ /\d/) ? m : '0' + m }
	s2.gsub(/\d+/) { |m| m + ((($`[-1..-1] != '.') && ($'[0..0] != '.')) ? '.0' : '') }
end

#Control the 4 states of the tool
def set_state(state)
	@state = state
	@state = STATE_EXECUTION if (@state > STATE_EXECUTION)
	if @state == STATE_DIRECTION
		@dir_base.ip.copy! @ip_origin
		@dir_base.pt = @pt_ref.clone
	elsif @state == STATE_ROTATION
		@dir_rot.ip.copy! @dir_base.ip
		@dir_rot.pt = @dir_base.pt.clone
	end	
	info_show
end

def onLButtonDown(flags, x, y, view)
	set_state @state + 1
end

#Handle Escape key
def onCancel(flag, view)
	return unless flag == 0
	return Sketchup.active_model.select_tool(nil) if (@state == STATE_ORIGIN)
	set_state @state - 1
end

def compute_origin
	@axesD = @normal.axes
	@tt = Geom::Transformation.axes(@origin, @axesD[0], @axesD[1], @axesD[2]) 	
	@color = get_color @normal
end

#OnMouseMove method for Tool
def onMouseMove(flags, x, y, view)
	case @state	
	when STATE_ORIGIN		#input Origin and Plane
		#view.lock_inference
		@ip_origin.pick view, x, y
		view.tooltip = @ip_origin.tooltip
		face = @ip_origin.face
		@origin = @ip_origin.position
		unless (flags == 4)		#unless Shift Key is down, to lock plane
			if (face)
				ph = view.pick_helper
				ph.do_pick x,y
				best = ph.best_picked
				if (best && best.typename == 'ComponentInstance')
					@normal = best.transformation * (face.normal)
				else
					@normal = @ip_origin.transformation * (face.normal)
				end	
				@face_color = (face.material) ? face.material.color : @gray_color
			else
				@normal = @normal_def
			end
		end	
		self.compute_origin
		view.invalidate
		return	
		
	when STATE_DIRECTION		#input direction
		direction_mousemove @dir_base, x, y, view, @axesD[0]
		
	when STATE_ROTATION			#placing the construction line
		direction_mousemove @dir_rot, x, y, view, @dir_base.vec
	end	
	view.invalidate	
end

#Draw method for tool
def draw(view)
	if (@state >= STATE_ORIGIN)
		#Compute the right scale to keep the protractor the same size
		size = view.pixels_to_model 1, @origin 	
		t = @tt * Geom::Transformation.scaling(size)
		
		#draw the origin
		@ip_origin.draw view

		#draw the protractor
		view.drawing_color = @color 
		pts = []
		@bigcircle.each {|pt| pts.push (t * pt)}
		view.draw GL_LINE_STRIP, pts
		@radius_bigcircle = @origin.distance pts[0]
		@pt_ref = pts[0]
		
		pts = []
		@graduations.each {|pt| pts.push (t * pt)}
		view.draw_lines pts	
		
		pts = []
		@inner_up.each {|pt| pts.push (t * pt)}
		view.draw GL_LINE_STRIP, pts
		
		pts = []
		@inner_down.each {|pt| pts.push (t * pt)}
		view.draw GL_LINE_STRIP, pts
	end
	
	#Draw the direction line
	if (@state >= STATE_DIRECTION)
		direction_draw @dir_base, view, false
	end
	
	#Draw the rotation line
	if (@state >= STATE_ROTATION)
		direction_draw @dir_rot, view, true
	end
end

#Input method on MOuse move for Direction structures @dir_base et @dir_rot
def direction_mousemove(dir, x, y, view, vecref) 
	dir.ip.pick view, x, y, @ip_origin
	view.tooltip = dir.ip.tooltip
	plane = [@origin, @normal]
	if (dir.ip.degrees_of_freedom <= dir.freedom)
		dir.pt = dir.ip.position.project_to_plane plane
		dir.angle_imposed = true
	else
		pickray = view.pickray x, y
		dir.pt = Geom.intersect_line_plane pickray, plane
		dir.angle_imposed  = false
	end	
	dir.vec = @origin.vector_to(dir.pt)
	angle = dir.vec.angle_between vecref
	v = dir.vec * vecref
	angle = 2 * Math::PI - angle if (v.length != 0 && @normal.samedirection?(v))
	
	if (dir.angle_imposed)
		direction_set_angle dir, angle
		dir.angle_imposed = true
	else
		factor = @origin.distance(dir.pt) / @radius_bigcircle
		direction_set_angle dir, adjust_angle(angle, factor)
	end	
	info_angle dir.angle
end

#Draw method for Direction structures @dir_base et @dir_rot
def direction_draw(dir, view, flgbox)
	dir.ip.draw view if dir.angle_imposed	
	view.line_stipple = "_"
	if (dir.angle_imposed && dir.pt != dir.ip.position)
		view.drawing_color = get_color dir.pt.vector_to(dir.ip.position)
		view.draw GL_LINES, [dir.pt, dir.ip.position]
	end	
	if (flgbox)
		pt1 = view.screen_coords @origin
		pt2 = view.screen_coords dir.pt
		clip_to_box pt1.x, pt1.y, pt2.x,pt2.y
		view.drawing_color = get_color dir.vec
		view.draw2d GL_LINES, [@ptbox[0], @ptbox[1]]
	else
		view.line_stipple = ""
		view.drawing_color = get_color dir.vec
		view.draw GL_LINE_STRIP, [@origin, dir.pt] if (@origin != dir.pt)
	end
end

#Force a particular value of the angle (from the VCB usually)
def direction_set_angle (dir, angle)
	d = @origin.distance dir.pt
	dir.angle = angle
	if (dir == @dir_base)
		v = dir.vec * @axesD[0]
		tdir = Geom::Transformation.rotation @origin, @normal, angle
		@tt = tdir * Geom::Transformation.axes(@origin, @axesD[0], @axesD[1], @axesD[2])
		dir.pt = @tt * @bigcircle[0]
	else
		dir.pt = Geom::Transformation.rotation(@origin, @normal, angle) * @pt_ref
	end
	dir.vec = @origin.vector_to dir.pt
	dir.pt = @origin.offset dir.vec, d
	dir.angle_imposed  = false
end

#Adjust angleto round value,depending on distance of input point to Origin
def adjust_angle(angle, factor)
	fac = factor.to_i + 1
	incr = 15.0
	tolerance = 1.0 / 2.0 / fac
	dangle = angle * 180 / Math::PI
	dangle = dangle.round if fac < 4
	a = dangle / incr
	around = a.round
	afinal = ((a - around).abs < tolerance) ? around * incr : dangle
	afinal * Math::PI / 180
end

#display angle value in the VCB
def info_angle(angle)
	dangle = angle * 180 / Math::PI
	dangle = dangle.modulo 360.0
	Sketchup.set_status_text sprintf("%3.1f ", dangle) + @msg_Degree, SB_VCB_VALUE
end

#display information in the Sketchup status bar
def info_show
	case @state
	when STATE_ORIGIN
		msg = @msg_Input_Origin
		label = @msg_PlaneAngle
	when STATE_DIRECTION
		msg = @msg_Input_Direction
		label = @msg_Direction
	when STATE_ROTATION
		msg = Traductor[MSG_Input_Rotation, @opname]
		label = @msg_Rotation
	when STATE_EXECUTION
		msg = Traductor[MSG_Input_Execution, @opname]
		label = @msg_Rotation
	end
	Sketchup.set_status_text msg
	Sketchup.set_status_text label, SB_VCB_LABEL
end
	
#Get color corresponding to a vector direction
def get_color vec
	if (vec == nil || vec.length == 0)
		colorname = "Black"
	elsif (vec.parallel? X_AXIS)
		colorname = "Red"
	elsif (vec.parallel? Y_AXIS)
		colorname = "Lawngreen"
	elsif (vec.parallel? Z_AXIS)
		colorname = "Blue"
	else
		colorname = "Black"
	end		
	color = adjust_color colorname
end

def adjust_color(colorname)
	prox = 25
	color = Sketchup::Color.new colorname
	return color if ((color.red - @face_color.red).abs > prox)
	return color if ((color.blue - @face_color.blue).abs > prox)
	return color if ((color.green - @face_color.green).abs > prox)
	color = Sketchup::Color.new 255 - color.red, 255 - color.green, 255 - color.blue 
end

#Compute the construction line for rotation vector - This is done in pixel
def clip_to_box (x1, y1, x2, y2)
	ay = (y1 - y2)
	ax = (x1 - x2)
	
	if (ay == 0)
		@ptbox[0].x = 0
		@ptbox[0].y = @ptbox[1].y = y1
		@ptbox[1].x = @vpwidth
	elsif (ax == 0)
		@ptbox[0].x = @ptbox[1].x = x1
		@ptbox[0].y = 0
		@ptbox[1].y = @vpheight
	else
		a = ay / ax
		i = 0
		v = y1 - a * x1						#left edge
		if (v >= 0 && v <= @vpheight)		
			@ptbox[i].x = 0
			@ptbox[i].y = v
			i += 1	
		end	
		v = y1 + a * (@vpwidth - x1)		#right edge
		if (v >= 0 && v <= @vpheight)
			@ptbox[i].x = @vpwidth
			@ptbox[i].y = v
			return if ((i += 1) > 1)
		end
		v = x1 - y1 / a 					#top edge
		if (v >= 0 && v <= @vpwidth)
			@ptbox[i].x = v
			@ptbox[i].y = 0
			return if ((i += 1) > 1)
		end	
		v = x1 + (@vpheight - y1) / a		#bottom edge
		if (v >= 0 && v <= @vpwidth)
			@ptbox[i].x = v
			@ptbox[i].y = @vpheight
		end
	end
end

end #Class ProtractorTool


end #End module LibProtractorTool